from .utils.other import API_URL, HEADERS_MAIN
from .handlers import fetch, requestExceptionHandler
from .classes.Exceptions import accountError, authDataError
from .classes.Objects import Points, Balances, Stats

#############################################################################
#     Module for working with your account on Portals Gift Marketplace.     #
#############################################################################

async def myPoints(authData: str = "") -> Points:
    """
    Retrieves the user's Portals Points information.

    Args:
        authData (str): The authentication data required for the API request.

    Returns:
        Points: An instance of Points containing the user's points information if the request is successful.

    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """
    URL = API_URL + "points/me"

    if authData == "":
        raise authDataError("portalsmp: myPoints(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")
    
    requestExceptionHandler(response, "myPoints")

    return Points(response.json())

async def myStats(authData: str = "") -> Stats:
    """
    Retrieves the user's statistics.

    Args:
        authData (str): The authentication data required for the API request.

    Returns:
        Stats: An instance of Stats containing the user's statistics if the request is successful.

    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """
    URL = API_URL + "users/profile/stats"

    if authData == "":
        raise authDataError("aportalsmp: myStats(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "myStats")

    return Stats(response.json())

async def myBalances(authData: str = "") -> Balances:
    """
    Retrieves the user's balances.

    Args:
        authData (str): The authentication data required for the API request.

    Returns:
        Balances: An instance of Balances containing the user's balances if the request is successful.

    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """
    URL = API_URL + "users/wallets/"

    if authData == "":
        raise authDataError("aportalsmp: myBalances(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "myBalances")

    return Balances(response.json())

async def withdrawPortals(amount: float = 0, wallet: str = "", authData: str = "") -> str:
    """
    Withdraw Portals from the user's wallet to an external address.

    Args:
        amount (float): The amount of Portals to withdraw.
        wallet (str): The external address to withdraw to.
        authData (str): The authentication data required for the API request.

    Returns:
        str: The ID of the withdrawal.

    Raises:
        Exception: If amount is not provided.
        Exception: If wallet is not provided.
        Exception: If authData is not provided.
        Exception: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "users/wallets/withdraw"

    if amount == 0:
        raise accountError("aportalsmp: withdrawPortals(): Error: amount is required")
    if not wallet:
        raise accountError("aportalsmp: withdrawPortals(): Error: wallet is required")
    if not authData:
        raise authDataError("aportalsmp: withdrawPortals(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    PAYLOAD = {
        "amount": str(amount),
        "external_address": wallet
        }
    
    response = await fetch(method="POST", url=URL, json=PAYLOAD, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "withdrawPortals")

    return response.json().get("id", None) if response.status_code == 200 else None


from .account import (
    myPoints,
    myStats,
    myBalances,
    withdrawPortals
)
from .auth import update_auth

from .classes.Objects import (
    PortalsGift,
    CollectionOffer,
    GiftOffer,
    Points,
    Stats,
    Balances,
    Filters,
    GiftsFloors,
    Collections,
    CollectionItem,
    Activity,
    MyActivity,
    SaleResult,
    Giveaway,
    GiveawayChannel,
    GiveawayPrize,
    GiveawayRequirements
)
from .classes.Exceptions import (
    authDataError,
    offerError,
    accountError,
    requestError,
    connectionError,
    floorsError,
    giftsError,
    tradingError
)

from .gifts import (
    giftsFloors,
    filterFloors,
    collections,
    search,
    marketActivity,
    myPortalsGifts,
    myActivity,
    buy,
    changePrice,
    bulkList,
    sale,
    getGiveaways,
    giveawayInfo,
    joinGiveaway,
    transferGifts,
    withdrawGifts
)
from .offers import (
    makeOffer,
    editOffer,
    cancelOffer,
    collectionOffer,
    editCollectionOffer,
    cancelCollectionOffer,
    topOffer,
    allCollectionOffers,
    myCollectionOffers,
    myReceivedOffers,
    myPlacedOffers,
)

from .utils.functions import (
    toShortName,
    convertForBuying,
    convertForListing
)

__all__ = [
    "myPoints",
    "myStats",
    "myBalances",
    "withdrawPortals",
    "update_auth",
    "giftsFloors",
    "filterFloors",
    "collections",
    "search",
    "marketActivity",
    "myPortalsGifts",
    "myActivity",
    "buy",
    "changePrice",
    "bulkList",
    "sale",
    "makeOffer",
    "editOffer",
    "cancelOffer",
    "collectionOffer",
    "editCollectionOffer",
    "cancelCollectionOffer",
    "topOffer",
    "allCollectionOffers",
    "myCollectionOffers",
    "myReceivedOffers",
    "myPlacedOffers",
    "toShortName",
    "convertForBuying",
    "convertForListing",
    "getGiveaways",
    "giveawayInfo",
    "joinGiveaway",
    "transferGifts",
    "withdrawGifts"
]



from .utils.functions import cap
from .utils.other import API_URL, HEADERS_MAIN
from .utils.collections_ids import collections_ids
from .classes.Exceptions import authDataError, offerError
from .classes.Objects import CollectionOffer, GiftOffer
from .handlers import fetch, requestExceptionHandler

#######################################################################
#     Module for working with offers on Portals Gift Marketplace.     #
#######################################################################

# =================== Gifts offers ===================

async def makeOffer(nft_id: str = "", offer_price: int | float = 0, expiration_days: int = 7, authData: str = "") -> None:
    """
    Creates an offer for a specified NFT.
    Args:
        nft_id (str): The unique identifier of the NFT to make an offer on.
        offer_price (int | float): The price of the offer.
        expiration_days (int): The number of days until the offer expires. Must be either 7 or 0 (no expiration). Default is 7.
        authData (str): The authentication data required for the API request.
    Returns:
        None: If the request is successful and the offer is created.
    Raises:
        offerError: If nft_id is not provided.
        offerError: If offer_price is not provided or is 0.
        offerError: If expiration_days is not 7 or 0.
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + "offers"

    if not nft_id:
        raise offerError("aportalsmp: makeOffer(): Error: nft_id is required")
    if offer_price == 0:
        raise offerError("aportalsmp: makeOffer(): Error: offer_price is required")
    if expiration_days not in [7, 0]:
        raise offerError("aportalsmp: makeOffer(): Error: expiration_days must be 7 or 0")
    if not authData:
        raise authDataError("aportalsmp: makeOffer(): authData is required.")


    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    PAYLOAD = {
        "offer": {
            "nft_id": nft_id,
            "offer_price": str(offer_price)
            }
    }

    if expiration_days == 7:
        PAYLOAD["offer"].update({"expiration_days": expiration_days})
    
    response = await fetch(method="POST", url=URL, json=PAYLOAD, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "makeOffer")
    
    return None

async def cancelOffer(offer_id: str = "", authData: str = "") -> None:
    """
    Cancels an offer with the given offer_id.

    Args:
        offer_id (str): The unique identifier of the offer to be canceled.
        authData (str): The authentication data required for the API request.

    Returns:
        None: If the request is successful and the offer is canceled.

    Raises:
        offerError: If offer_id is not provided.
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + "offers/" + f"{offer_id}" + "/cancel"

    if not offer_id:
        raise offerError("aportalsmp: cancelOffer(): Error: offer_id is required")
    if authData == "":
        raise authDataError("aportalsmp: cancelOffer(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="POST", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "cancelOffer")

    return None

async def editOffer(offer_id: str = "", new_price: int | float = 0, authData: str = "") -> None:
    """
    Edit existing offer price.
    Args:
        offer_id (str): The unique identifier of the offer to be edited.
        new_price (int | float): The new price to set for the offer.
        authData (str): The authentication data required for the API request.
    Returns:
        None: If the request is successful and the offer is edited.
    Raises:
        offerError: If offer_id is not provided.
        offerError: If new_price is not provided or is less than 0.5.
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + "offers/" + f"{offer_id}"

    if not offer_id:
        raise offerError("aportalsmp: editOffer(): Error: offer_id is required")
    if type(new_price) not in [float, int] or new_price < 0.5:
        raise offerError("aportalsmp: editOffer(): Error: new_price must be a number >= 0.5")
    if not authData:
        raise authDataError("aportalsmp: editOffer(): Error: authData is required")

    PAYLOAD = {
        "amount": str(float(new_price))
    }

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="PATCH", url=URL, json=PAYLOAD, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "editOffer")
   
    return None

# =================== Collection offers ===================

async def collectionOffer(gift_name: str = "", amount: float | int = 0, expiration_days: int = 7, max_nfts: int = 1, authData: str = "") -> None:
    """
    Make an offer for collection.
    Args:
        gift_name (str): The name of the gift collection.
        amount (float | int): Price in TON you want to offer.
        expiration_days (int): The number of days until the offer expires. Must be either 7 or 0 (no expiration). Default is 7.
        max_nfts (int): The maximum number of NFTs that can be purchased with this offer. Default is 1.
        authData (str): The authentication data required for the API request.
    Returns:
        None: If the request is successful and the offer is created.
    Raises:
        offerError: If gift_name is not provided.
        offerError: If amount is not provided or is less than or equal to 0.
        offerError: If max_nfts is not provided or is less than or equal to 0.
        authDataError: If authData is not provided.
        offerError: If expiration_days is not 0 or 7.
        requestError: If the API request fails.
    """

    URL = API_URL + "collection-offers/"

    if not gift_name:
        raise offerError("aportalsmp: collectionOffer(): Error: gift_name is required")
    if amount <= 0.0:
        raise offerError("aportalsmp: collectionOffer(): Error: amount is required")
    if max_nfts <= 0:
        raise offerError("aportalsmp: collectionOffer(): Error: max_nfts is required")
    if not authData:
        raise authDataError("aportalsmp: collectionOffer(): Error: authData is required")
    if expiration_days not in [0,7]:
        raise offerError("aportalsmp: collectionOffer(): Error: expiration_days must be 0 (no expiration) or 7 (7 days)")

    gift_name = cap(gift_name)

    ID = collections_ids.get(gift_name, None)

    if ID is None:
        raise offerError("aportalsmp: collectionOffer(): Error: gift_name is invalid")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    PAYLOAD = {
        "amount": str(amount),
        "collection_id": ID,
        "expiration_days": expiration_days,
        "max_nfts": max_nfts
    }

    response = await fetch(method="POST", url=URL, headers=HEADERS, json=PAYLOAD, impersonate="chrome110")

    requestExceptionHandler(response, "collectionOffer")
    
    return None

async def editCollectionOffer(amount: float | int = 0, offer_id: str = "", max_nfts: int = 1, authData: str = "") -> None:
    """
    Edit existing collection offer's price.
    Args:
        amount (float | int): The new price to set for the collection offer.
        offer_id (str): The unique identifier of the collection offer to be edited.
        max_nfts (int): The maximum number of NFTs that can be purchased with this offer. Default is 1.
        authData (str): The authentication data required for the API request.
    Returns:
        None: If the request is successful and the collection offer is edited.
    Raises:
        offerError: If amount is not provided or is less than 0.5.
        offerError: If offer_id is not provided.
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + "collection-offers/update"

    if not offer_id:
        raise offerError("aportalsmp: editCollectionOffer(): Error: offer_id is required")
    if type(amount) not in [float, int] or amount < 0.5:
        raise offerError("aportalsmp: editCollectionOffer(): Error: amount must be number >= 0.5")
    if not authData:
        raise authDataError("aportalsmp: editCollectionOffer(): Error: authData is required")

    PAYLOAD = {
        "amount": str(float(amount)),
        "id": offer_id,
        "max_nfts": max_nfts
    }

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="POST", url=URL, headers=HEADERS, json=PAYLOAD, impersonate="chrome110")

    requestExceptionHandler(response, "editCollectionOffer")

    return None

async def cancelCollectionOffer(offer_id: str = "", authData: str = "") -> None:
    """
    Cancels a collection offer with the given offer_id.

    Args:
        offer_id (str): The unique identifier of the collection offer to be canceled.
        authData (str): The authentication data required for the API request.
    Returns:
        None: If the request is successful and the collection offer is canceled.
    Raises:
        offerError: If offer_id is not provided.
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + "collection-offers/cancel"

    if not offer_id:
        raise offerError("aportalsmp: cancelCollectionOffer(): Error: offer_id is required")

    if not authData:
        raise authDataError("aportalsmp: cancelCollectionOffer(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    PAYLOAD = {
        "id": offer_id
    }

    response = await fetch(method="POST", url=URL, headers=HEADERS, json=PAYLOAD, impersonate="chrome110")

    requestExceptionHandler(response, "cancelCollectionOffer")

    return None

async def allCollectionOffers(gift_name: str = "", authData: str = "") -> list[CollectionOffer]:
    """
    Retrieves all collection offers for a specific gift collection.
    Args:
        gift_name (str): The name of the gift collection.
        authData (str): The authentication data required for the API request.
    Returns:
        list: A list of CollectionOffer objects if the request is successful.
    Raises:
        offerError: If gift_name is not provided.
        offerError: If gift_name is invalid.
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + "collection-offers/"

    if not gift_name:
        raise offerError("aportalsmp: allCollectionOffers(): Error: gift_name is required")
    
    gift_name = cap(gift_name)
    ID = collections_ids.get(gift_name, None)

    if ID is None:
        raise offerError("aportalsmp: allCollectionOffers(): Error: gift_name is invalid")
    if not authData:
        raise authDataError("aportalsmp: allCollectionOffers(): Error: authData is required")
    
    URL += f"{ID}/all"
    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "allCollectionOffers")

    return [CollectionOffer(offer) for offer in response.json()] if len(response.json()) > 0 else []

async def topOffer(gift_name: str = "", authData: str = "") -> CollectionOffer | None:
    """
    Retrieves the top offer for a specified gift collection.

    Args:
        gift_name (str): The name of the gift collection.
        authData (str): The authentication data required for the API request.
    Returns:
        CollectionOffer | None: A CollectionOffer object if the request is successful and a top offer exists.
    Raises:
        offerError: If gift_name is not provided or is invalid.
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + "collection-offers/"

    try:
        ID = collections_ids[cap(gift_name)]
    except:
        raise offerError("aportalsmp: topOffer(): Error: gift_name is invalid")

    if authData == "":
        raise authDataError("aportalsmp: topOffer(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    URL += f"{ID}/top"

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "topOffer")

    return CollectionOffer(response.json()) if response.status_code == 200 and len(response.json()) > 0 else None

# ==================== My offers ===================

async def myReceivedOffers(offset: int = 0, limit: int = 20, authData: str = "") -> list[GiftOffer]:
    """
    Retrieves the offers received by the user.

    Args:
        offset (int): The pagination offset. Defaults to 0.
        limit (int): The maximum number of results to return. Defaults to 20.
        authData (str): The authentication data required for the API request.
    Returns:
        list: A list of GiftOffer objects if the request is successful.
    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + f"offers/received?offset={offset}&limit={limit}"

    if authData == "":
        raise authDataError("aportalsmp: myReceivedOffers(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "myReceivedOffers")

    return [GiftOffer(offer["offer"]) for offer in response.json()["top_offers"]] if "top_offers" in response.json() and len(response.json()["top_offers"]) > 0 else []

async def myCollectionOffers(authData: str = "") -> list[CollectionOffer]:
    """
    Retrieves the collection offers placed by the user.

    Args:
        authData (str): The authentication data required for the API request.
    Returns:
        list: A list of CollectionOffer objects if the request is successful.
    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + "collection-offers/"
    if authData == "":
        raise authDataError("aportalsmp: myCollectionOffers(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "myCollectionOffers")

    return [CollectionOffer(offer) for offer in response.json()] if len(response.json()) > 0 else []

async def myPlacedOffers(offset: int = 0, limit: int = 20, authData: str = "") -> list[GiftOffer]:
    """
    Retrieves the offers placed by the user.

    Args:
        offset (int): The pagination offset. Defaults to 0.
        limit (int): The maximum number of results to return. Defaults to 20.
        authData (str): The authentication data required for the API request.

    Returns:
        list: A list of GiftOffer objects if the request is successful.

    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + f"offers/placed?offset={offset}&limit={limit}"

    if authData == "":
        raise authDataError("aportalsmp: myPlacedOffers(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "myPlacedOffers")

    return [GiftOffer(offer) for offer in response.json()["offers"]] if "offers" in response.json() and len(response.json()["offers"]) > 0 else []


from pyrogram import Client
from pyrogram.raw.functions.messages import RequestAppWebView
from pyrogram.raw.types import InputBotAppShortName, InputUser
from pyrogram.raw.functions.users import GetUsers
from urllib.parse import unquote
from .classes.Exceptions import authDataError
import os

#################################
#     Authentication module     #
#################################

async def update_auth(api_id: int|str = "", api_hash: str = "", session_string: str = "", session_path: str = "", session_name: str = "account") -> str:
    """
    Updates Telegram authData for Portals API using Pyrogram.

    Args:
        api_id (int|str)
        api_hash (str)
        session_string (str) - if you already have a session string, you can pass it here (without passing api_id and api_hash).
        session_path (str, optional) - path where the session file will be saved. Default is current working directory.
        session_name (str, optional) - name of the session file (without extension). Default is "account".

    Returns:
        str: new authData
    """
    if not session_path:
        session_path = os.getcwd()
    if api_id and api_hash and not session_string:
        async with Client(f"{session_name}", api_id=api_id, api_hash=api_hash, workdir=session_path) as client:
            peer = await client.resolve_peer("portals")
            user_full = await client.invoke(GetUsers(id=[peer]))
            bot_raw = user_full[0]
            bot = InputUser(user_id=bot_raw.id, access_hash=bot_raw.access_hash)
            bot_app = InputBotAppShortName(bot_id=bot, short_name="market")
            web_view = await client.invoke(
                RequestAppWebView(
                    peer=peer,
                    app=bot_app,
                    platform="desktop",
                )
            )
            initData = unquote(web_view.url.split('tgWebAppData=', 1)[1].split('&tgWebAppVersion', 1)[0])
            return f"tma {initData}"
    elif session_string:
        async with Client(f"{session_name}", session_string=session_string, workdir=session_path) as client:
            peer = await client.resolve_peer("portals")
            user_full = await client.invoke(GetUsers(id=[peer]))
            bot_raw = user_full[0]
            bot = InputUser(user_id=bot_raw.id, access_hash=bot_raw.access_hash)
            bot_app = InputBotAppShortName(bot_id=bot, short_name="market")
            web_view = await client.invoke(
                RequestAppWebView(
                    peer=peer,
                    app=bot_app,
                    platform="desktop",
                )
            )
            initData = unquote(web_view.url.split('tgWebAppData=', 1)[1].split('&tgWebAppVersion', 1)[0])
            return f"tma {initData}"
    else:
        raise authDataError("aportalsmp: update_auth(): You must provide either api_id and api_hash or a session_string.")



from typing import Optional

from .classes.Exceptions import requestError, connectionError
from curl_cffi.requests import AsyncSession


###########################
#     Handlers module     #
###########################

async def fetch(method: str = "GET",
                url: str = "",
                headers: dict = None,
                json: dict = None,
                timeout: int = 15,
                impersonate: str = "chrome110",
                proxies: Optional[dict] = None):
    """
    Fetches a URL with optional headers and JSON payload.
    
    Args:
        url (str): The URL to fetch.
        headers (dict, optional): Headers to include in the request.
        json (dict, optional): JSON payload to include in the request.
        method (str, optional): HTTP method to use. Defaults to "GET".
        impersonate (str, optional): User agent to impersonate. Defaults to "chrome110".

    Returns:
        Response object from the requests library.
    """
    async with AsyncSession(impersonate=impersonate, timeout=timeout, proxies=proxies) as session:
        try:
            response = await session.request(method=method, url=url, headers=headers, json=json)
            return response
        except:
            raise connectionError(
                f"aportalsmp: fetch(): Error: Connection reset after {timeout} seconds. Please check your internet connection or try again later.")


def requestExceptionHandler(response, func_name):
    if not (200 <= response.status_code < 300):
        try:
            responseJson = response.json()
            message = responseJson.get("message")
            if message:
                raise requestError(
                    f"aportalsmp: {func_name}(): Error: status_code: {response.status_code}, message: {message}")
            else:
                raise requestError(
                    f"aportalsmp: {func_name}(): Error: status_code: {response.status_code}, json: {responseJson}")
        except ValueError:
            if len(response.text) > 300:
                raise requestError(
                    f"aportalsmp: {func_name}(): Error: status_code: {response.status_code}. Response text is too long to display (likely raw HTML).")
            else:
                raise requestError(
                    f"aportalsmp: {func_name}(): Error: status_code: {response.status_code}, response_text: {response.text}")



from typing import Optional

from .utils.other import API_URL, HEADERS_MAIN, SORTS
from .utils.functions import toShortName, cap, listToURL, activityListToURL
from .utils.collections_ids import collections_ids
from .handlers import requestExceptionHandler, fetch
from .classes.Exceptions import authDataError, floorsError, giftsError, tradingError
from .classes.Objects import GiftsFloors, Filters, Collections, PortalsGift, Activity, MyActivity, SaleResult, Giveaway, GiveawayRequirements
from urllib.parse import quote_plus

######################################################################
#     Module for working with gifts on Portals Gift Marketplace.     #
######################################################################

# ================ Floors ================

async def giftsFloors(authData: str = "") -> GiftsFloors:
    """
    Retrieves the floor prices for all gift collections (short names only).

    Args:
        authData (str): The authentication data required for the API request.

    Returns:
        GiftsFloors: An instance of GiftsFloors containing the floor prices of collections if the request is successful.

    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails.
    """

    URL = API_URL + "collections/floors"

    if authData == "":
        raise authDataError("aportalsmp: giftsFloors(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "giftsFloors")

    return GiftsFloors(response.json()['floorPrices'])

async def filterFloors(gift_name: str = "", authData: str = "") -> Filters:
    """
    Retrieves the floor prices of models/backdrops/symbols for a specific gift collection.
    Args:
        gift_name (str): The name of the gift collection.
        authData (str): The authentication data required for the API request.
    Returns:
        Filters: An instance of Filters containing the floor prices of models/backdrops/symbols for the specified gift collection if the request is successful.
    Raises:
        authDataError: If authData is not provided.
        floorsError: If gift_name is not provided or is not a string.
        floorsError: If gift_name is not a valid string.
        requestError: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "collections/filters"

    if not authData:
        raise authDataError("aportalsmp: filters(): Error: authData is required")
    if not gift_name:
        raise floorsError("aportalsmp: filters(): Error: gift_name is required")
    if type(gift_name) == str:
        gift_name = toShortName(gift_name)
    if type(gift_name) != str:
        raise floorsError("aportalsmp: filters(): Error: gift_name must be a string")

    URL += f"?short_names={gift_name}"

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "filterFloors")

    return Filters(response.json()['floor_prices'][gift_name]) if "floor_prices" in response.json() and response.json()['floor_prices'].get(gift_name, None) is not None else None

async def collections(limit: int = 100, authData: str = "") -> Collections:
    """
    Retrieves a list of collections and their floors, supply, daily volume etc from the marketplace.

    Args:
        limit (int): The maximum number of results to return. Defaults to 100.
        authData (str): The authentication data required for the API request.

    Returns:
        Collections: An instance of Collections containing the list of collections and their details if the request is successful.

    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "collections" + f"?limit={limit}"

    if authData == "":
        raise authDataError("aportalsmp: collections(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "collections")

    return Collections(response.json()['collections'])

# ================ Gifts ================

async def search(sort: str = "price_asc",
                 offset: int = 0,
                 limit: int = 20,
                 gift_name: str | list = "",
                 model: str | list = "",
                 backdrop: str | list = "",
                 symbol: str | list = "",
                 min_price: int = 0,
                 max_price: int = 100000,
                 exclude_bundled: bool = True,
                 premarket_status: str = "all",
                 authData: str = "",
                 proxies: Optional[dict] = None) -> list[PortalsGift]:
    """
    Search for gifts with various filters and sorting options.

    Args:
        sort (str): The sorting method for the results. Options include "latest", "price_asc", "price_desc", 
            "gift_id_asc", "gift_id_desc", "model_rarity_asc", "model_rarity_desc". Defaults to "price_asc".
        offset (int): The pagination offset (limit*page). Defaults to 0.
        limit (int): The maximum number of results to return. Defaults to 20.
        gift_name (str | list): The name or list of names of gifts to filter.
        model (str | list): The model or list of models to filter.
        backdrop (str | list): The backdrop or list of backdrops to filter.
        symbol (str | list): The symbol or list of symbols to filter.
        min_price (int): The minimum price of the gifts to filter. Defaults to 0.
        max_price (int): The maximum price of the gifts to filter. Defaults to 100000.
        authData (str): The authentication data required for the API request.
    Returns:
        list[PortalsGift]: A list of PortalsGift objects containing the search results.
    Raises:
        authDataError: If authData is not provided.
        giftsError: If max_price is less than min_price or if min_price and max_price are not integers.
        giftsError: If gift_name, model, backdrop, or symbol are not strings or lists.
        giftsError: If sort is not one of the valid options.
        giftsError: If premarket_status is not one of the valid options.
        requestError: If the API request fails.
    
    НОВЫЕ ПАРАМЕТРЫ:
        exclude_bundled (bool): Whether to exclude bundled gifts. True or False. Defaults to True.
        premarket_status (str): Premarket status filter. Options: "all", "only_premarket", "without_premarket". Defaults to "all".
    """

    if sort not in SORTS:
        raise giftsError(f"aportalsmp: search(): Error: sort must be one of these options: {list(SORTS.keys())}")
    
    PREMARKET_STATUSES = ["all", "only_premarket", "without_premarket"]
    if premarket_status not in PREMARKET_STATUSES:
        raise giftsError(f"aportalsmp: search(): Error: premarket_status must be one of these options: {PREMARKET_STATUSES}")

    URL = API_URL + "nfts/" + "search?" + f"offset={offset}" + f"&limit={limit}" + f"{SORTS[sort]}" 

    try:
        min_price = int(min_price)
        max_price = int(max_price)
    except:
        raise giftsError("aportalsmp: search(): Error: min_price and max_price must be integers")
    
    if max_price < 100000:
        URL += f"&min_price={min_price}&max_price={max_price}"

    if authData == "":
        raise authDataError("aportalsmp: search(): Error: authData is required")
    if max_price < min_price:
        raise giftsError("aportalsmp: search(): Error: max_price must be greater than min_price")

    if gift_name:
        if type(gift_name) == str:
            # ТУт нейм в айди конверт
            gift_id = collections_ids.get(cap(gift_name))
            if gift_id:
                URL += f"&collection_ids={gift_id}"
            else:
                raise giftsError(f"aportalsmp: search(): Error: gift_name '{gift_name}' not found in collections_ids")
        elif type(gift_name) == list:
            # Тут список неймов в айди конверт
            gift_ids = []
            for name in gift_name:
                gift_id = collections_ids.get(cap(name))
                if gift_id:
                    gift_ids.append(gift_id)
                else:
                    raise giftsError(f"aportalsmp: search(): Error: gift_name '{name}' not found in collections_ids")
            URL += f"&collection_ids={','.join(gift_ids)}"
        else:
            raise giftsError("aportalsmp: search(): Error: gift_name must be a string or list")
    if model:
        if type(model) == str:
            URL += f"&filter_by_models={quote_plus(cap(model))}"
        elif type(model) == list:
            URL += f"&filter_by_models={listToURL(model)}"
        else:
            raise giftsError("aportalsmp: search(): Error: model must be a string or list")
    if backdrop:
        if type(backdrop) == str:
            URL += f"&filter_by_backdrops={quote_plus(cap(backdrop))}"
        elif type(backdrop) == list:
            URL += f"&filter_by_backdrops={listToURL(backdrop)}"
        else:
            raise giftsError("aportalsmp: search(): Error: backdrop must be a string or list")
    if symbol:
        if type(symbol) == str:
            URL += f"&filter_by_symbols={quote_plus(cap(symbol))}"
        elif type(symbol) == list:
            URL += f"&filter_by_symbols={listToURL(symbol)}"
        else:
            raise giftsError("aportalsmp: search(): Error: symbol must be a string or list")

    URL += "&status=listed"
    URL += f"&exclude_bundled={str(exclude_bundled).lower()}"
    URL += f"&premarket_status={premarket_status}"

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110", proxies=proxies)

    requestExceptionHandler(response, "search")

    return [PortalsGift(gift) for gift in response.json()["results"]]

async def marketActivity(sort: str = "latest", offset: int = 0, limit: int = 20, activityType: str | list = "", gift_name: str | list= "", model: str | list = "", backdrop: str | list = "", symbol: str | list = "", min_price: int = 0, max_price: int = 100000, authData: str = "") -> list[Activity]:
    """
    Retrieves market activity with various filters and sorting options.

    Args:
        sort (str): The sorting method for the results. Options include "latest", "price_asc", "price_desc", 
            "gift_id_asc", "gift_id_desc", "model_rarity_asc", "model_rarity_desc". Defaults to "latest".
        offset (int): The pagination offset (limit*page). Defaults to 0.
        limit (int): The maximum number of results to return. Defaults to 20.
        activityType (str): The type of activity to filter by. Options are "buy", "listing", "price_update", 
            "offer", or an empty string for no filter.
        gift_name (str | list): The name or list of names of gifts to filter.
        model (str | list): The model or list of models to filter.
        backdrop (str | list): The backdrop or list of backdrops to filter.
        symbol (str | list): The symbol or list of symbols to filter.
        min_price (int): The minimum price of the gifts to filter. Defaults to 0.
        max_price (int): The maximum price of the gifts to filter. Defaults to 100000.
        authData (str): The authentication data required for the API request.

    Returns:
        list[Activity]: A list of Activity objects containing the market activity results.

    Raises:
        authDataError: If authData is not provided.
        giftsError: If max_price is less than min_price or if min_price and max_price are not integers.
        giftsError: If activityType is not a valid string or list.
        giftsError: If gift_name, model, backdrop, or symbol are not strings or lists.
        requestError: If the API request fails.
    """

    URL = API_URL + "market/actions/" + f"?offset={offset}" + f"&limit={limit}" + f"{SORTS[sort]}"

    try:
        min_price = int(min_price)
        max_price = int(max_price)
    except:
        raise giftsError("aportalsmp: marketActivity(): Error: min_price and max_price must be integers")

    if max_price < 100000:
        URL += f"&min_price={min_price}&max_price={max_price}"

    if authData == "":
        raise authDataError("aportalsmp: marketActivity(): Error: authData is required")
    if max_price < min_price:
        raise giftsError("aportalsmp: marketActivity(): Error: max_price must be greater than min_price")
    if type(activityType) == str and activityType.lower() not in ["", "buy", "listing", "price_update", "offer"]:
        raise giftsError("aportalsmp: marketActivity(): Error: activityType may be empty, buy, listing, offer or price_update only.")
    if type(activityType) == list:
        activityType = activityListToURL(activityType)

    if gift_name:
        if type(gift_name) == str:
            URL += f"&filter_by_collections={quote_plus(cap(gift_name))}"
        elif type(gift_name) == list:
            URL += f"&filter_by_collections={listToURL(gift_name)}"
        else:
            raise giftsError("aportalsmp: marketActivity(): Error: gift_name must be a string or list")
    if model:
        if type(model) == str:
            URL += f"&filter_by_models={quote_plus(cap(model))}"
        elif type(model) == list:
            URL += f"&filter_by_models={listToURL(model)}"
        else:
            raise giftsError("aportalsmp: marketActivity(): Error: model must be a string or list")
    if backdrop:
        if type(backdrop) == str:
            URL += f"&filter_by_backdrops={quote_plus(cap(backdrop))}"
        elif type(backdrop) == list:
            URL += f"&filter_by_backdrops={listToURL(backdrop)}"
        else:
            raise giftsError("aportalsmp: marketActivity(): Error: backdrop must be a string or list")
    if symbol:
        if type(symbol) == str:
            URL += f"&filter_by_symbols={quote_plus(cap(symbol))}"
        elif type(symbol) == list:
            URL += f"&filter_by_symbols={listToURL(symbol)}"
        else:
            raise giftsError("aportalsmp: marketActivity(): Error: symbol must be a string or list")
        
    if activityType:
        URL += f"&action_types={activityType}"

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "marketActivity")

    return [Activity(activity) for activity in response.json()['actions']]

async def getGiveaways(offset: int = 0, limit: int = 20, authData: str = "") -> list[Giveaway]:
    """
    Retrieves a list of giveaways from the marketplace.

    Args:
        offset (int): The pagination offset (limit*(page+1)). Defaults to 0.
        limit (int): The maximum number of results to return. Defaults to 20.
        authData (str): The authentication data required for the API request.
    Returns:
        list[Giveaway]: A list of Giveaway objects containing the giveaway details.
    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "giveaways/" + f"?offset={offset}" + f"&limit={limit}" + f"&status=active"

    if not authData:
        raise authDataError("aportalsmp: getGiveaways(): Error: authData is required")
    if type(offset) != int or type(limit) != int:
        raise giftsError("aportalsmp: getGiveaways(): Error: offset and limit must be integers")
    
    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "getGiveaways")

    return [Giveaway(giveaway) for giveaway in response.json()['giveaways']] if "giveaways" in response.json() else []

async def giveawayInfo(giveaway_id: str = "", authData: str = "") -> GiveawayRequirements:
    """
    Retrieves the requirements for a specific giveaway.

    Args:
        giveaway_id (str): The unique identifier of the giveaway.
        authData (str): The authentication data required for the API request.

    Returns:
        GiveawayRequirements: An instance of GiveawayRequirements containing the giveaway requirements if the request is successful.

    Raises:
        authDataError: If authData is not provided.
        giftsError: If giveaway_id is not provided or is not a string.
        requestError: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "giveaways/" + f"{giveaway_id}/" + "requirements"

    if not giveaway_id:
        raise giftsError("aportalsmp: giveawayInfo(): Error: giveaway_id is required")
    if type(giveaway_id) != str:
        raise giftsError("aportalsmp: giveawayInfo(): Error: giveaway_id must be a string")
    if not authData:
        raise authDataError("aportalsmp: giveawayInfo(): Error: authData is required")
    
    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "giveawayInfo")

    return GiveawayRequirements(response.json()) if response.status_code == 200 else None

async def joinGiveaway(giveaway_id: str = "", authData: str = "") -> None:
    """
    Joins a giveaway with the specified giveaway_id.

    Args:
        giveaway_id (str): The unique identifier of the giveaway to join.
        authData (str): The authentication data required for the API request.

    Returns:
        None: if the request is successful, likely a 204 no content response from the API.

    Raises:
        authDataError: If authData is not provided.
        giftsError: If giveaway_id is not provided or is not a string.
        requestError: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "giveaways/" + f"{giveaway_id}/" + "join"

    if not giveaway_id:
        raise giftsError("aportalsmp: joinGiveaway(): Error: giveaway_id is required")
    if type(giveaway_id) != str:
        raise giftsError("aportalsmp: joinGiveaway(): Error: giveaway_id must be a string")
    if not authData:
        raise authDataError("aportalsmp: joinGiveaway(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="POST", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "joinGiveaway")

    if response.status_code == 204:
        return None

# ================ My ================

async def myPortalsGifts(offset: int = 0, limit: int = 20, listed: bool = True, authData: str = "") -> list[PortalsGift]:
    """
    Retrieves a list of the user's owned Portal gifts.

    Args:
        offset (int): The offset for pagination.
        limit (int): The maximum number of items to return.
        listed (bool): If True, only gifts listed for sale are shown. Defaults to True.
        authData (str): The authentication data required for the API request.

    Returns:
        list[PortalsGift]: A list of PortalsGift objects representing the user's owned gifts.
    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "nfts/" + "owned?" + f"offset={offset}" + f"&limit={limit}"

    if authData == "":
        raise authDataError("aportalsmp: myPortalsGifts(): Error: authData is required")
    if listed == True:
        URL += "&status=listed"
    else:
        URL += "&status=unlisted"
    
    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "myPortalsGifts")

    return [PortalsGift(nft) for nft in response.json()['nfts']] if "nfts" in response.json() else []

async def myActivity(offset: int = 0, limit: int = 20, authData: str = "") -> list[MyActivity]:
    """
    Retrieves the user's activity on the marketplace.

    Args:
        offset (int): The pagination offset (limit*page). Defaults to 0.
        limit (int): The maximum number of results to return. Defaults to 20.
        authData (str): The authentication data required for the API request.

    Returns:
        list[MyActivity]: A list of MyActivity objects containing the user's activity results.

    Raises:
        authDataError: If authData is not provided.
        requestError: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "users/actions/" + f"?offset={offset}" + f"&limit={limit}"

    if authData == "":
        raise authDataError("aportalsmp: myActivity(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    response = await fetch(method="GET", url=URL, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "myActivity")

    return [MyActivity(action) for action in response.json()['actions']]

async def transferGifts(nft_ids: list = [], username: str = "", anonymous: bool = False, authData: str = "") -> None:
    """
    Transfers gifts to a specified user inside the marketplace.

    Args:
        nft_ids (list): A list of NFT IDs to be transferred.
        username (str): The username of the recipient to whom the gifts will be transferred.
        anonymous (bool): If True, the transfer will be anonymous. Defaults to False.
        authData (str): The authentication data required for the API request.

    Returns:
        None: if the request is successful, likely a 204 no content response from the API.

    Raises:
        authDataError: If authData is not provided.
        giftsError: If nft_ids is not a non-empty list or if username is not provided.
        requestError: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "nfts/transfer-gifts"

    if not authData:
        raise authDataError("aportalsmp: transferGifts(): Error: authData is required")
    if type(nft_ids) != list or len(nft_ids) == 0:
        raise giftsError("aportalsmp: transferGifts(): Error: nft_ids must be a non-empty list")
    if not username:
        raise giftsError("aportalsmp: transferGifts(): Error: username is required")
    if anonymous not in [0, 1, True, False]:
        raise giftsError("aportalsmp: transferGifts(): Error: anonymous must be a boolean value (True or False)")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    PAYLOAD = {
        "nft_ids": nft_ids,
        "recipient": username,
        "anonymous": bool(anonymous)
    }

    response = await fetch(method="POST", url=URL, json=PAYLOAD, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "transferGifts")

    return None

# ================ Trading ================

async def buy(nft_id: str = "", price: int|float = 0, authData: str = ""):
    """
    Buys a gift with the given nft_id at the given price.

    Args:
        nft_id (str): The unique identifier of the NFT to be bought.
        price (int|float): The price at which the NFT should be bought.
        authData (str): The authentication data required for the API request.

    Returns:
        dict: JSON-ответ API. Например:
            {
               "total_requested":1,
               "total_purchased":0,
               "total_failed":1,
               "total_spent":"0",
               "purchase_results":[
                  {
                     "id":"fea0382a-dc66-4173-8a31-680b0ac55671",
                     "status":"failed",
                     "reason":"INSUFFICIENT_BALANCE",
                     "price":"2.32",
                     "error_message":"Insufficient balance to purchase this NFT"
                  }
               ]
            } 
    Raises:
        authDataError: If authData is not provided.
        tradingError: If nft_id is not provided or price is not a positive number.
        requestError: If the API request fails or returns a non-200 status code.
    """
    URL = API_URL + "nfts"

    if authData == "":
        raise authDataError("aportalsmp: buy(): Error: authData is required")
    if not nft_id:
        raise tradingError("aportalsmp: buy(): Error: nft_id is required")
    if type(price) not in [int, float] or price <= 0:
        raise tradingError("aportalsmp: buy(): Error: price error")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    nfts = [{"id": nft_id, "price": str(price)}]
    '''
    {"nft_details":[{"id":"aaaa8eb4-deac-4ba2-aa5c-ea79c73f0d5b","price":"1.85"}]}
    '''

    PAYLOAD = {
        "nft_details": nfts
    }

    response = await fetch(method="POST", url=URL, json=PAYLOAD, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "buy")

    return response.json()

async def changePrice(nft_id: str = "", price: float = 0, authData: str = "") -> None:
    """
    Updates the price of a specified NFT.

    Args:
        nft_id (str): The unique identifier of the NFT for which the price is being updated.
        price (float): The new price to set for the NFT.
        authData (str): The authentication data required for the API request.

    Returns:
        None: if the request is successful, likely a 204 no content response from the API.

    Raises:
        authDataError: If authData is not provided.
        tradingError: If nft_id is not provided or price is not a positive number.
        requestError: If the API request fails or returns a non-200 status code.
    """

    URL = API_URL + "nfts/" + f"{nft_id}/" + "list"

    if not nft_id:
        raise tradingError("aportalsmp: changePrice(): Error: nft_id is required")
    if type(price) not in [int, float] or price <= 0:
        raise tradingError("aportalsmp: changePrice(): Error: price is required")
    if authData == "":
        raise authDataError("aportalsmp: changePrice(): Error: authData is required")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    PAYLOAD = {
        "price": str(price)
    }

    response = await fetch(method="POST", url=URL, json=PAYLOAD, headers=HEADERS, impersonate="chrome110")
    
    requestExceptionHandler(response, "changePrice")

    return None


async def bulkList(nfts: list = [], authData: str = "") -> SaleResult:
    """
    Lists multiple NFTs for sale in bulk.

    Args:
        nfts (list): A non-empty list of dictionaries, each containing 'nft_id', and 'price' as keys.
        authData (str): The authentication data required for the API request.

    Returns:
        SaleResult: An instance of SaleResult containing the response from the API if the request is successful.

    Raises:
        authDataError: If authData is not provided.
        tradingError: If nfts is not a non-empty list.
        requestError: If the API request fails or returns a non-200 status code.
    """

    URL = API_URL + "nfts/bulk-list"

    if authData == "":
        raise authDataError("aportalsmp: bulkList(): Error: authData is required")
    if type(nfts) != list or len(nfts) == 0:
        raise tradingError("aportalsmp: bulkList(): Error: nfts must be a non-empty list")

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    PAYLOAD = {
        "nft_prices": nfts
    }

    response = await fetch(method="POST", url=URL, json=PAYLOAD, headers=HEADERS, impersonate="chrome110")
    
    requestExceptionHandler(response, "bulkList")

    return SaleResult(response.json())

async def sale(nft_id: str = "", price: int|float = 0,authData: str = "") -> SaleResult:
    """
    Lists a single NFT for sale.

    Args:
        nft_id (str): The unique identifier of the NFT to be listed.
        price (int|float): The price at which the NFT should be listed for sale.
        authData (str): The authentication data required for the API request.

    Returns:
        SaleResult: An instance of SaleResult containing the response from the API if the request is successful.

    Raises:
        authDataError: If authData is not provided.
        tradingError: If nft_id is not provided or price is not a positive number.
        requestError: If the API request fails or returns a non-200 status code.
    """

    URL = API_URL + "nfts/bulk-list"

    if authData == "":
        raise authDataError("aportalsmp: sale(): Error: authData is required")
    if not nft_id:
        raise tradingError("aportalsmp: sale(): Error: nft_id is required")
    if price == 0 or type(price) not in [int, float]:
        raise tradingError("aportalsmp: sale(): Error: price error")

    nfts = [{"nft_id": nft_id, "price": str(price)}]

    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    PAYLOAD = {
        "nft_prices": nfts
    }

    response = await fetch(method="POST", url=URL, json=PAYLOAD, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "sale")

    return SaleResult(response.json())

async def withdrawGifts(nft_ids: list = [], authData: str = "") -> None:
    """
    [NOT TESTED]
    Withdraws multiple / single NFT(s) from the marketplace.

    Args:
        nft_ids (list): A list of NFT IDs to withdraw. If a single NFT is to be withdrawn, pass a list with one item.
        authData (str): The authentication data required for the API request.

    Returns:
        None: if the request is successful, likely a 204 no content response from the API.

    Raises:
        authDataError: If authData is not provided.
        tradingError: If nft_ids is not a non-empty list.
        requestError: If the API request fails or returns a non-200 status code.
    """

    URL = API_URL + "nfts/withdraw"

    if not authData:
        raise authDataError("aportalsmp: withdrawGifts(): Error: authData is required")
    if type(nft_ids) != list or len(nft_ids) == 0:
        raise tradingError("aportalsmp: withdrawGifts(): Error: nft_ids must be a non-empty list")
    
    HEADERS = {**HEADERS_MAIN, "Authorization": authData}

    PAYLOAD = {
        "gift_ids": nft_ids
    }

    response = await fetch(method="POST", url=URL, json=PAYLOAD, headers=HEADERS, impersonate="chrome110")

    requestExceptionHandler(response, "withdrawGifts")

    return None
